@@ -5,8 +5,12 @@ import tempfile
 import unittest
 
 import numpy
-import pyfits
+try:
+    from astropy.io import fits
+except:
+    import pyfits as fits
 import cpl
+cpl.Recipe.memory_mode = 0
 
 recipe_name = 'rtest'
 raw_tag = 'RRRECIPE_DOCATG_RAW'
@@ -23,8 +27,8 @@ class RecipeTestCase(CplTestCase):
         self.recipe.temp_dir = self.temp_dir
         self.recipe.tag = raw_tag
         self.image_size = (16, 16)
-        self.raw_frame = pyfits.HDUList([
-                pyfits.PrimaryHDU(numpy.random.random_integers(0, 65000,
+        self.raw_frame = fits.HDUList([
+                fits.PrimaryHDU(numpy.random.random_integers(0, 65000,
                                                                self.image_size))])
         self.raw_frame[0].header.update('HIERARCH ESO DET DIT', 0.0)
         self.raw_frame[0].header.update('HIERARCH ESO PRO CATG', raw_tag)
@@ -267,15 +271,15 @@ class RecipeCalib(RecipeTestCase):
         '''[TAB] completition. 
         This requires to have   the __dir__() method working.
         '''
-        self.assertEqual(self.recipe.calib.__dir__(), 
-                         [ f.tag for f in self.recipe.calib ])
+        self.assertEqual(set(self.recipe.calib.__dir__()),
+                         set(f.tag for f in self.recipe.calib))
 
 class RecipeExec(RecipeTestCase):
     def setUp(self):
         RecipeTestCase.setUp(self)
-        self.flat_frame = pyfits.HDUList([
-                pyfits.PrimaryHDU(numpy.random.random_integers(0, 65000,
-                                                               self.image_size))])
+        self.flat_frame = fits.HDUList([
+                fits.PrimaryHDU(numpy.random.random_integers(0, 65000,
+                                                             self.image_size))])
 
     def test_frames_keyword_dict(self):
         '''Raw and calibration frames specified as keyword dict'''
@@ -283,9 +287,13 @@ class RecipeExec(RecipeTestCase):
         res = self.recipe(raw = {'RRRECIPE_DOCATG_RAW': self.raw_frame },
                           calib = { 'FLAT':self.flat_frame })
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
         self.assertTrue(abs(self.raw_frame[0].data 
                             - res.THE_PRO_CATG_VALUE[0].data).max() == 0)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_frames_keyword_calib(self):
         '''Raw frame specified as keyword, calibration frame set in recipe'''
@@ -293,7 +301,11 @@ class RecipeExec(RecipeTestCase):
         self.recipe.calib.FLAT = self.flat_frame
         res = self.recipe({'RRRECIPE_DOCATG_RAW':self.raw_frame})
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_frames_tag_keyword(self):
         '''The 'tag' parameter'''
@@ -301,22 +313,34 @@ class RecipeExec(RecipeTestCase):
         self.recipe.calib.FLAT = self.flat_frame
         res = self.recipe(self.raw_frame, tag = raw_tag)
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_frames_tag_attribute(self):
         '''The 'tag' attribute'''
         self.recipe.tag = raw_tag
         res = self.recipe(self.raw_frame)
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_frames_one_element_input_list(self):
         '''Use 1-element list as input'''
         # --> we want a list back'''
         res = self.recipe([self.raw_frame])
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertFalse(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertFalse(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
         self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, list))
+        try:
+            res.THE_PRO_CATG_VALUE[0].close()
+        except:
+            pass
 
     def test_frames_many_element_input_list(self):
         '''Use multiple files as input'''
@@ -324,7 +348,11 @@ class RecipeExec(RecipeTestCase):
         # assumed to be a 'master', and we get back a plain frame'''
         res = self.recipe([self.raw_frame, self.raw_frame])
         self.assertTrue(isinstance(res, cpl.Result))
-        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+        self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_output_dir_attribute(self):
         '''Write an output dir specified as attribute'''
@@ -337,8 +365,12 @@ class RecipeExec(RecipeTestCase):
                          'rtest.fits')
         self.assertTrue(os.path.isdir(output_dir))
         self.assertTrue(os.path.isfile(res.THE_PRO_CATG_VALUE))
-        hdu = pyfits.open(res.THE_PRO_CATG_VALUE)
-        self.assertTrue(isinstance(hdu, pyfits.HDUList))
+        hdu = fits.open(res.THE_PRO_CATG_VALUE)
+        self.assertTrue(isinstance(hdu, fits.HDUList))
+        try:
+            hdu.close()
+        except:
+            pass
 
     def test_output_dir_keyword(self):
         '''Write an output dir specified as call keyword arg'''
@@ -352,14 +384,42 @@ class RecipeExec(RecipeTestCase):
         self.assertEqual(os.path.basename(res.THE_PRO_CATG_VALUE), 
                          'rtest.fits')
         self.assertTrue(os.path.isfile(res.THE_PRO_CATG_VALUE))
-        hdu = pyfits.open(res.THE_PRO_CATG_VALUE)
-        self.assertTrue(isinstance(hdu, pyfits.HDUList))
+        hdu = fits.open(res.THE_PRO_CATG_VALUE)
+        self.assertTrue(isinstance(hdu, fits.HDUList))
+        try:
+            hdu.close()
+        except:
+            pass
+
+    def test_param_default(self):
+        '''Test default parameter settings'''
+        res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
+        self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'].strip(),
+                         self.recipe.param.stropt.default or '')
+        self.assertEqual(res[0].header['HIERARCH ESO QC BOOLOPT'],
+                         self.recipe.param.boolopt.default)
+        self.assertEqual(res[0].header['HIERARCH ESO QC INTOPT'],
+                         self.recipe.param.intopt.default)
+        self.assertEqual(res[0].header['HIERARCH ESO QC FLOATOPT'],
+                         self.recipe.param.floatopt.default)
+        self.assertEqual(res[0].header['HIERARCH ESO QC ENUMOPT'],
+                         self.recipe.param.enumopt.default)
+        self.assertEqual(res[0].header['HIERARCH ESO QC RANGEOPT'],
+                         self.recipe.param.rangeopt.default)
+        try:
+            res.close()
+        except:
+            pass
 
     def test_param_keyword_dict(self):
         '''Parameter handling via keyword dict'''
         res = self.recipe(self.raw_frame, 
                           param = { 'stropt':'more' }).THE_PRO_CATG_VALUE
         self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'more')
+        try:
+            res.close()
+        except:
+            pass
 
     def test_param_keyword_dict_wrong(self):
         '''Parameter handling via keyword dict'''
@@ -371,14 +431,26 @@ class RecipeExec(RecipeTestCase):
         self.recipe.param.stropt = 'more'
         res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
         self.assertEqual(res[0].header['HIERARCH ESO QC STROPT'], 'more')
+        try:
+            res.close()
+        except:
+            pass
 
     def test_param_delete(self):
         '''Delete a parameter in a second run after setting it'''
         self.recipe.param.intopt = 123
         res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
+        try:
+            res.close()
+        except:
+            pass
         del self.recipe.param.intopt
         res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
         self.assertEqual(res[0].header['HIERARCH ESO QC INTOPT'], 2)
+        try:
+            res.close()
+        except:
+            pass
 
     def test_param_overwrite(self):
         '''Overwrite the recipe setting param via via keyword arg'''
@@ -401,18 +473,42 @@ class RecipeExec(RecipeTestCase):
         self.assertEqual(res[0].header['HIERARCH ESO QC FLOATOPT'], -0.25)
         self.assertEqual(res[0].header['HIERARCH ESO QC ENUMOPT'], 'third')
         self.assertEqual(res[0].header['HIERARCH ESO QC RANGEOPT'], 0.125)
+        try:
+            res.close()
+        except:
+            pass
         
+    def test_disabled(self):
+        '''Parameter with CLI disabled'''
+        self.assertFalse(self.recipe.param.disabled.enabled[0])
+        self.assertTrue(self.recipe.param.intopt.enabled[0])
+#        self.recipe.param.disabled = 0.2
+#        res = self.recipe(self.raw_frame)
+#        self.assertEqual(res[0].header['HIERARCH ESO QC DISABLED'], 0.2)
+#        try:
+#            res.close()
+#        except:
+#            pass
+
     def test_environment_setting(self):
         '''Additional environment parameter via recipe setting'''
         self.recipe.env['TESTENV'] = 'unkk'
         res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
         self.assertEqual(res[0].header['HIERARCH ESO QC TESTENV'], 'unkk')
+        try:
+            res.close()
+        except:
+            pass
 
     def test_environment_keyword(self):
         '''Additional environment parameter via recipe call keyword'''
         res = self.recipe(self.raw_frame, 
                           env = {'TESTENV':'kknu'}).THE_PRO_CATG_VALUE
         self.assertEqual(res[0].header['HIERARCH ESO QC TESTENV'], 'kknu')
+        try:
+            res.close()
+        except:
+            pass
 
     def test_error(self):
         '''Error handling'''
@@ -430,7 +526,7 @@ class RecipeExec(RecipeTestCase):
                                        threaded = True))
         for i, res in enumerate(results):
             # check if we got the correct type
-            self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, pyfits.HDUList))
+            self.assertTrue(isinstance(res.THE_PRO_CATG_VALUE, fits.HDUList))
             # check if we have the correct parameter
             self.assertEqual(res.THE_PRO_CATG_VALUE[0].header[
                     'HIERARCH ESO QC INTOPT'], i)
@@ -443,6 +539,10 @@ class RecipeExec(RecipeTestCase):
             # check that the data were moved correctly
             self.assertTrue(abs(self.raw_frame[0].data 
                                 - res.THE_PRO_CATG_VALUE[0].data).max() < 1e-6)
+            try:
+                res.THE_PRO_CATG_VALUE.close()
+            except:
+                pass
 
     def test_error_parallel(self):
         '''Error handling in parallel execution'''
@@ -458,6 +558,10 @@ class RecipeExec(RecipeTestCase):
         res = self.recipe(self.raw_frame)
         key = 'DATAMD5'
         md5sum = res.THE_PRO_CATG_VALUE[0].header[key]
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
         self.assertNotEqual(md5sum, 'Not computed')
         self.assertEqual(len(md5sum), 
                          len('9d123996fa9a7bda315d07e063043454'))
@@ -469,6 +573,10 @@ class RecipeExec(RecipeTestCase):
         res = self.recipe(self.raw_frame)
         key = 'HIERARCH ESO PRO REC1 CAL1 DATAMD5'
         md5sum = res.THE_PRO_CATG_VALUE[0].header[key]
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
         self.assertNotEqual(md5sum, 'Not computed')
         self.assertEqual(len(md5sum), 
                          len('9d123996fa9a7bda315d07e063043454'))
@@ -498,15 +606,21 @@ class RecipeRes(RecipeTestCase):
         RecipeTestCase.setUp(self)
         self.res = self.recipe(self.raw_frame)
 
+    def tearDown(self):
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
+
     def test_attribute(self):
         '''The result as an attribute'''
         self.assertTrue(isinstance(self.res.THE_PRO_CATG_VALUE, 
-                                   pyfits.HDUList))
+                                   fits.HDUList))
 
     def test_dict(self):
         '''The result as an attribute'''
         self.assertTrue(isinstance(self.res['THE_PRO_CATG_VALUE'], 
-                                   pyfits.HDUList))
+                                   fits.HDUList))
 
     def test_len(self):
         '''Length of the result'''
@@ -516,7 +630,7 @@ class RecipeRes(RecipeTestCase):
         '''Iterate over the result'''
         for tag, hdu in self.res:
             self.assertEqual(tag, 'THE_PRO_CATG_VALUE')
-            self.assertTrue(isinstance(hdu, pyfits.HDUList))
+            self.assertTrue(isinstance(hdu, fits.HDUList))
 
 class RecipeEsorex(CplTestCase):
     def setUp(self):
@@ -579,10 +693,31 @@ class RecipeEsorex(CplTestCase):
         filename = os.path.join(dirname, filename)
         logging.getLogger('cpl').info(log_msg)
         self.assertTrue(os.path.exists(filename))
-        log_content = open(filename).read()
+        logfile = open(filename)
+        log_content = logfile.read()
+        logfile.close()
         self.assertTrue(log_msg in log_content)
         self.assertTrue('INFO' in log_content)
 
+    def test_esorex_log_off(self):
+        '''Switch the logfile off after writing something'''
+        dirname = os.path.join(self.temp_dir, 'log')
+        filename = 'python-cpl_off.log'
+        log_msg = 'Esorex convienence log'
+        os.mkdir(dirname)
+        cpl.esorex.log.dir = dirname
+        cpl.esorex.log.filename = 'python-cpl_debug.log'
+        cpl.esorex.log.level = 'debug'
+        logging.getLogger('cpl').debug(log_msg)
+        cpl.esorex.log.filename = filename
+        cpl.esorex.log.level = 'off'
+        logging.getLogger('cpl').debug(log_msg)
+        filename = os.path.join(dirname, filename)
+        logfile = open(filename)
+        log_content = logfile.read()
+        logfile.close()
+        self.assertEqual(len(log_content), 0)
+
 class RecipeLog(RecipeTestCase):
     def setUp(self):
         RecipeTestCase.setUp(self)
@@ -611,7 +746,11 @@ class RecipeLog(RecipeTestCase):
         '''Injection of CPL messages into the python logging system'''
         self.handler.clear()
         logging.getLogger().setLevel(logging.DEBUG)
-        self.recipe(self.raw_frame)
+        res = self.recipe(self.raw_frame)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
         # check that the logs are not empty
         self.assertNotEqual(len(self.handler.logs), 0)
@@ -634,7 +773,12 @@ class RecipeLog(RecipeTestCase):
         '''Filtering INFO messages'''
         self.handler.clear()
         logging.getLogger('cpl.rtest').setLevel(logging.INFO)
-        self.recipe(self.raw_frame)
+        res = self.recipe(self.raw_frame)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
+
         # check that the logs are not empty
         self.assertNotEqual(len(self.handler.logs), 0)
 
@@ -642,7 +786,12 @@ class RecipeLog(RecipeTestCase):
         '''Filtering WARN messages'''
         self.handler.clear()
         logging.getLogger('cpl.rtest').setLevel(logging.WARN)
-        self.recipe(self.raw_frame)
+        res = self.recipe(self.raw_frame)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
+
         # check that the logs are not empty
         self.assertNotEqual(len(self.handler.logs), 0)
 
@@ -651,21 +800,33 @@ class RecipeLog(RecipeTestCase):
         # There is no error msg written by the recipe, so it should be empty.
         self.handler.clear()
         logging.getLogger('cpl.rtest').setLevel(logging.ERROR)
-        self.recipe(self.raw_frame)
+        res = self.recipe(self.raw_frame)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
         self.assertEqual(len(self.handler.logs), 0)
 
     def test_logging_common(self):
         '''Log name specification on recipe call'''
         self.handler.clear()
         self.other_handler.clear()
-        self.recipe(self.raw_frame, logname = 'othername')
+        res = self.recipe(self.raw_frame, logname = 'othername')
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
         self.assertNotEqual(len(self.other_handler.logs), 0)
 
     def test_logging_multiline(self):
         '''Multiple lines in messages'''
         self.handler.clear()
         logging.getLogger('cpl.rtest').setLevel(logging.INFO)
-        self.recipe(self.raw_frame)
+        res = self.recipe(self.raw_frame)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
         # check that the multi line log sequence appears
         multiline = 0
         tag = 'multiline#'
@@ -696,6 +857,10 @@ class RecipeLog(RecipeTestCase):
         self.assertTrue(isinstance(res.log.warning[0], str))
         # Check that there were no error messages
         self.assertEqual(len(res.log.error), 0)
+        try:
+            res.THE_PRO_CATG_VALUE.close()
+        except:
+            pass
 
     def test_error(self):
         '''"log" attribute of the CplError object'''
@@ -730,12 +895,18 @@ class ProcessingInfo(RecipeTestCase):
         self.recipe.param.floatopt = -0.25
         self.recipe.param.enumopt = 'third'
         self.recipe.param.rangeopt = 0.125
-        self.recipe.calib.FLAT = pyfits.HDUList([
-                pyfits.PrimaryHDU(numpy.random.random_integers(0, 65000,
+        self.recipe.calib.FLAT = fits.HDUList([
+                fits.PrimaryHDU(numpy.random.random_integers(0, 65000,
                                                           self.image_size))])
         self.res = self.recipe(self.raw_frame).THE_PRO_CATG_VALUE
         self.pinfo = cpl.dfs.ProcessingInfo(self.res)
 
+    def tearDown(self):
+        try:
+            self.res.close()
+        except:
+            pass
+
     def test_param(self):
         '''Parameter information'''
         self.assertEqual(len(self.pinfo.param), len(self.recipe.param))
